@@ -3,11 +3,17 @@
 from __future__ import print_function
 import struct
 import numpy
+import io
 from .types import LenType, DataType, MeasurementType
 from .types import ScanType, UnitType, DataType
-from .types import Offsets
+from .types import Offsets, ExifTags
 from .utils import convert_wl, convert_attr_name
 from sys import stderr
+try:
+    import PIL
+    from PIL import Image
+except ImportError:
+    PIL = None
 
 
 class WDFReader(object):
@@ -29,7 +35,7 @@ class WDFReader(object):
     `BKXL`: ? TODO
     `WXCS`: ? TODO
     `WXIS`: ? TODO
-    `WHTL`: An embeded image (?)
+    `WHTL`: Whilte light image
 
     Following the block name, there are two indicators:
     Block uid: int32
@@ -50,11 +56,11 @@ class WDFReader(object):
     count (int) : Numbers of experiments (same type), can be smaller than capacity
     spectral_units (int) : Unit of spectra, see unit_types
     xlist_type (int) : See unit_types
-    xlist_units (int) : See unit_types
+    xlist_unit (int) : See unit_types
     xlist_length (int): Size for the xlist
     xdata (numpy.array): x-axis data
     ylist_type (int): Same as xlist_type
-    ylist_units (int): Same as xlist_units
+    ylist_unit (int): Same as xlist_unit
     ylist_length (int): Same as xlist_length
     ydata (numpy.array): y-data, possibly not used
     point_per_spectrum (int): Should be identical to xlist_length
@@ -66,7 +72,7 @@ class WDFReader(object):
                         # TODO types?
     """
 
-    def __init__(self, file_name, quiet=False):
+    def __init__(self, file_name, quiet=True):
         try:
             self.file_obj = open(str(file_name), "rb")
         except IOError:
@@ -74,15 +80,15 @@ class WDFReader(object):
         # Initialize the properties for the wdfReader class
         self.title = ""
         self.username = ""
-        self.measurement_type = ""
-        self.scan_type = ""
+        self.measurement_type = None
+        self.scan_type = None
         self.laser_length = None
         self.count = None
-        self.spectral_units = ""
+        self.spectral_unit = None
         self.xlist_type = None
-        self.xlist_units = ""
+        self.xlist_unit = None
         self.ylist_type = None
-        self.ylist_units = ""
+        self.ylist_unit = None
         self.point_per_spectrum = None
         self.data_origin_count = None
         self.capacity = None
@@ -92,29 +98,30 @@ class WDFReader(object):
         self.ylist_length = 0
         self.accumulation_count = None
         self.block_info = {}    # each key has value (uid, offset, size)
-        self.quiet = quiet      # if need to output infomation
         self.is_completed = False
         # Parse the header section in the wdf file
-        self._locate_all_blocks()
-        self._treat_block_data("WDF1")
-        self._treat_block_data("DATA")
-        self._treat_block_data("XLST")
-        self._treat_block_data("YLST")
-        self._treat_block_data("ORGN")
-        self._treat_block_data("WMAP")
+        self.__locate_all_blocks()
+        # Parse individual blocks
+        self.__treat_block_data("WDF1")
+        self.__treat_block_data("DATA")
+        self.__treat_block_data("XLST")
+        self.__treat_block_data("YLST")
+        self.__treat_block_data("ORGN")
+        self.__treat_block_data("WMAP")
+        self.__treat_block_data("WHTL")
 
         # Reshape spectra after reading mapping information
-        self._reshape_spectra()
+        self.__reshape_spectra()
         # self._parse_wmap()
 
         # Finally print the information
-        if not self.quiet:
+        if not quiet:
             self.print_info()
 
     def close(self):
         self.file_obj.close()
 
-    def _get_type_string(self, attr, data_type):
+    def __get_type_string(self, attr, data_type):
         """Get the enumerated-data_type as string
         """
         val = getattr(self, attr)  # No error checking
@@ -124,7 +131,7 @@ class WDFReader(object):
         else:
             return data_type(val).name
 
-    def _read_type(self, type, size=1):
+    def __read_type(self, type, size=1):
         """ Unpack struct data for certain type
         """
         if type in ["int16", "int32", "int64", "float", "double"]:
@@ -141,25 +148,25 @@ class WDFReader(object):
         else:
             raise ValueError("Unknown data length format!")
 
-    def _locate_single_block(self, pos):
+    def __locate_single_block(self, pos):
         """Get block information starting at pos
         """
         self.file_obj.seek(pos)
         block_name = self.file_obj.read(0x4).decode("ascii")
         if len(block_name) < 4:
             raise EOFError
-        block_uid = self._read_type("int32")
-        block_size = self._read_type("int64")
+        block_uid = self.__read_type("int32")
+        block_size = self.__read_type("int64")
         return block_name, block_uid, block_size
 
-    def _locate_all_blocks(self):
+    def __locate_all_blocks(self):
         """Get information for all data blocks and store them inside self.block_info
         """
         curpos = 0
         finished = False
         while not finished:
             try:
-                block_name, block_uid, block_size = self._locate_single_block(
+                block_name, block_uid, block_size = self.__locate_single_block(
                     curpos)
                 # print(block_name, block_uid, block_size)
                 self.block_info[block_name] = (block_uid, curpos, block_size)
@@ -167,20 +174,22 @@ class WDFReader(object):
             except (EOFError, UnicodeDecodeError):
                 finished = True
 
-    def _treat_block_data(self, block_name):
+    def __treat_block_data(self, block_name):
         """Get data according to specific block name
         """
         if block_name not in self.block_info.keys():
             print("Block name {0} not present in current measurement".
                   format(block_name), file=stderr)
             return
+        # parse individual blocks with names
         actions = {
             "WDF1": ("_parse_header", ()),
             "DATA": ("_parse_spectra", ()),
             "XLST": ("_parse_xylist", ("X")),
             "YLST": ("_parse_xylist", ("Y")),
             "ORGN": ("_parse_orgin_list", ()),
-            "WMAP": ("_parse_wmap", ())
+            "WMAP": ("_parse_wmap", ()),
+            "WHTL": ("_parse_img", ()),
         }
         func_name, val = actions[block_name]
         getattr(self, func_name)(*val)
@@ -193,42 +202,43 @@ class WDFReader(object):
         self.file_obj.seek(0)   # return to the head
         # Must make the conversion under python3
         block_ID = self.file_obj.read(Offsets.block_id).decode("ascii")
-        block_UID = self._read_type("int32")
-        block_len = self._read_type("int64")
+        block_UID = self.__read_type("int32")
+        block_len = self.__read_type("int64")
         # First block must be "WDF1"
-        if (block_ID != "WDF1") or (block_UID != 0 and block_UID != 1) \
-           or (block_len != 512):
+        if (block_ID != "WDF1") \
+           or (block_UID != 0 and block_UID != 1) \
+           or (block_len != Offsets.data_block):
             raise ValueError("The wdf file format is incorrect!")
         # TODO what are the digits in between?
 
         # The keys from the header
         self.file_obj.seek(Offsets.measurement_info)  # space
-        self.point_per_spectrum = self._read_type("int32")
-        self.capacity = self._read_type("int64")
-        self.count = self._read_type("int64")
+        self.point_per_spectrum = self.__read_type("int32")
+        self.capacity = self.__read_type("int64")
+        self.count = self.__read_type("int64")
         # If count < capacity, this measurement is not completed
         self.is_completed = (self.count == self.capacity)
-        self.accumulation_count = self._read_type("int32")
-        self.ylist_length = self._read_type("int32")
-        self.xlist_length = self._read_type("int32")
-        self.data_origin_count = self._read_type("int32")
-        self.application_name = self._read_type("utf8", 24)  # Must be "WiRE"
+        self.accumulation_count = self.__read_type("int32")
+        self.ylist_length = self.__read_type("int32")
+        self.xlist_length = self.__read_type("int32")
+        self.data_origin_count = self.__read_type("int32")
+        self.application_name = self.__read_type("utf8", 24)  # Must be "WiRE"
         for i in range(4):
-            self.application_version[i] = self._read_type("int16")
-        self.scan_type = self._read_type("int32")
-        self.measurement_type = self._read_type("int32")
+            self.application_version[i] = self.__read_type("int16")
+        self.scan_type = ScanType(self.__read_type("int32"))
+        self.measurement_type = MeasurementType(self.__read_type("int32"))
         # For the units
         self.file_obj.seek(Offsets.spectral_info)
-        self.spectral_units = self._read_type("int32")
-        self.laser_length = convert_wl(self._read_type("float"))
+        self.spectral_unit = UnitType(self.__read_type("int32"))
+        self.laser_length = convert_wl(self.__read_type("float"))  # in nm
         # Username and title
         self.file_obj.seek(Offsets.file_info)
-        self.username = self._read_type("utf8",
-                                        Offsets.usr_name -
-                                        Offsets.file_info)
-        self.title = self._read_type("utf8",
-                                     Offsets.data_block -
-                                     Offsets.usr_name)
+        self.username = self.__read_type("utf8",
+                                         Offsets.usr_name -
+                                         Offsets.file_info)
+        self.title = self.__read_type("utf8",
+                                      Offsets.data_block -
+                                      Offsets.usr_name)
 
     def _parse_xylist(self, dir):
         """Get information from XLST or YLST blocks
@@ -240,9 +250,9 @@ class WDFReader(object):
         offset = Offsets.block_data
         self.file_obj.seek(pos + offset)
         setattr(self, "{0}list_type".format(dir.lower()),
-                self._read_type("int32"))
-        setattr(self, "{0}list_units".format(dir.lower()),
-                self._read_type("int32"))
+                DataType(self.__read_type("int32")))
+        setattr(self, "{0}list_unit".format(dir.lower()),
+                UnitType(self.__read_type("int32")))
         size = getattr(self, "{0}list_length".format(dir.lower()))
         if size == 0:           # Possibly not started
             raise ValueError("{0}-List possibly not initialized!".
@@ -257,7 +267,7 @@ class WDFReader(object):
         """Get information from DATA block
         """
         if end == -1:           # take all spectra
-            end = self.count-1
+            end = self.count - 1
         if (start not in range(self.count)) or (end not in range(self.count)):
             raise ValueError("Wrong start and end indices of spectra!")
         if start > end:
@@ -297,9 +307,9 @@ class WDFReader(object):
 
         for i in range(self.data_origin_count):
             self.file_obj.seek(curpos)
-            p1 = self._read_type("int32")
-            p2 = self._read_type("int32")
-            s = self._read_type("utf8", 0x10)
+            p1 = self.__read_type("int32")
+            p2 = self.__read_type("int32")
+            s = self.__read_type("utf8", 0x10)
             # First index: is the list x, or y pos?
             self.origin_list_header[i][0] = (p1 >> 31 & 0b1) == 1
             # Second: Data type of the row
@@ -310,14 +320,16 @@ class WDFReader(object):
             self.origin_list_header[i][3] = s
             # Last: the actual data
             # array = numpy.empty(self.count)
-            array = numpy.array([self._read_type("double")
+            array = numpy.array([self.__read_type("double")
                                  for i in range(self.count)])
             self.origin_list_header[i][4] = array
             # Set self.xpos or self.ypos
-            if self.origin_list_header[i][1] == DataType.X:
+            if self.origin_list_header[i][1] == DataType.Spatial_X:
                 self.xpos = array
-            elif self.origin_list_header[i][1] == DataType.Y:
+                self.xpos_unit = self.origin_list_header[i][2]
+            elif self.origin_list_header[i][1] == DataType.Spatial_Y:
                 self.ypos = array
+                self.ypos_unit = self.origin_list_header[i][2]
             else:
                 pass
             curpos += list_increment
@@ -333,23 +345,23 @@ class WDFReader(object):
             return
 
         self.file_obj.seek(pos + Offsets.wmap_origin)
-        x_start = self._read_type("float")
+        x_start = self.__read_type("float")
         if not numpy.isclose(x_start, self.xpos[0], rtol=1e-4):
             raise ValueError("WMAP Xpos is not same as in ORGN!")
-        y_start = self._read_type("float")
+        y_start = self.__read_type("float")
         if not numpy.isclose(y_start, self.ypos[0], rtol=1e-4):
             raise ValueError("WMAP Ypos is not same as in ORGN!")
-        unknown1 = self._read_type("float")
-        x_pad = self._read_type("float")
-        y_pad = self._read_type("float")
-        unknown2 = self._read_type("float")
+        unknown1 = self.__read_type("float")
+        x_pad = self.__read_type("float")
+        y_pad = self.__read_type("float")
+        unknown2 = self.__read_type("float")
         # for i in range(9):
         #     print(i, self._read_type("int32"))
         #     self.file_obj.seek(pos + Offsets.wmap_origin + i * 4)
         #     print(i, self._read_type("float"))
         # self.file_obj.seek(pos + Offsets.wmap_wh)
-        self.spectra_w = self._read_type("int32")
-        self.spectra_h = self._read_type("int32")
+        spectra_w = self.__read_type("int32")
+        spectra_h = self.__read_type("int32")
         # print(self.spectra_w, self.spectra_h)
         # print(len(self.xpos), len(self.ypos))
         if len(self.xpos) > 1:
@@ -358,18 +370,93 @@ class WDFReader(object):
                 raise ValueError("WMAP Xpad is not same as in ORGN!")
         # If ypos smaller than spectra_w,
         # posssibly not even finished single line scan
-        if len(self.ypos) >= self.spectra_w:
-            if self.spectra_h > 1:
-                loc = self.spectra_w
+        if len(self.ypos) >= spectra_w:
+            if spectra_h > 1:
+                loc = spectra_w
             else:
                 loc = 1
             if not numpy.isclose(y_pad, self.ypos[loc] - self.ypos[0],
                                  rtol=1e-4):
                 raise ValueError("WMAP Ypad is not same as in ORGN!")
+        self.map_shape = (spectra_w, spectra_h)
+        self.map_info = dict(x_start=x_start,
+                             y_start=y_start,
+                             x_pad=x_pad,
+                             y_pad=y_pad,
+                             x_span=spectra_w * x_pad,
+                             y_span=spectra_h * y_pad,
+                             x_unit=self.xpos_unit,
+                             y_unit=self.ypos_unit)
 
-        self.map_info = (x_start, y_start, x_pad, y_pad)
+    def _parse_img(self):
+        """Extract the white-light JPEG image
+           The size of while-light image is coded in its EXIF
+           Use PIL to parse the EXIF information
+        """
+        try:
+            uid, pos, size = self.block_info["WHTL"]
+        except KeyError:
+            print("The wdf file does not contain an image",
+                  file=stderr)
+            return
 
-    def _reshape_spectra(self):
+        # Read the bytes. `self.img` is a wrapped IO object mimicking a file
+        self.file_obj.seek(pos + Offsets.jpeg_header)
+        img_bytes = self.file_obj.read(size - Offsets.jpeg_header)
+        self.img = io.BytesIO(img_bytes)
+        # Handle image dimension if PIL is present
+        if PIL is not None:
+            pil_img = Image.open(self.img)
+            exif_header = dict(pil_img.getexif())
+            try:
+                # Get the width and height of image
+                w_ = exif_header[ExifTags.FocalPlaneXResolution]
+                h_ = exif_header[ExifTags.FocalPlaneYResolution]
+                x_org_, y_org_ = exif_header[ExifTags.FocalPlaneXYOrigins]
+                # The dimensions (width, height)
+                # with unit `img_dimension_unit`
+                self.img_dimensions = numpy.array([w_[0] / w_[1],
+                                                   h_[0] / h_[1]])
+                # Origin of image is at upper right corner
+                self.img_origins = numpy.array([x_org_[0] / x_org_[1],
+                                                y_org_[0] / y_org_[1]])
+                # Default is microns (5)
+                self.img_dimension_unit = UnitType(
+                    exif_header[ExifTags.FocalPlaneResolutionUnit])
+                # Give the box for cropping
+                # Following the PIL manual
+                # (left, upper, right, lower)
+                self.img_cropbox = self.__calc_crop_box()
+
+            except KeyError:
+                print("Some keys in white light image header cannot be read!",
+                      file=stderr)
+        return
+
+    def __calc_crop_box(self):
+        """Helper function to calculate crop box
+        """
+        def _proportion(x, minmax, pixels):
+            """Get proportional pixels"""
+            min, max = minmax
+            return int(pixels * (x - min) / (max - min))
+
+        pil_img = PIL.Image.open(self.img)
+        w_, h_ = self.img_dimensions
+        x0_, y0_ = self.img_origins
+        pw = pil_img.width
+        ph = pil_img.height
+        map_xl = self.xpos.min()
+        map_xr = self.xpos.max()
+        map_yt = self.ypos.min()
+        map_yb = self.ypos.max()
+        left = _proportion(map_xl, (x0_, x0_ + w_), pw)
+        right = _proportion(map_xr, (x0_, x0_ + w_), pw)
+        top = _proportion(map_yt, (y0_, y0_ + h_), ph)
+        bottom = _proportion(map_yb, (y0_, y0_ + h_), ph)
+        return (left, top, right, bottom)
+
+    def __reshape_spectra(self):
         """Reshape spectra into w * h * self.point_per_spectrum
         """
         if not self.is_completed:
@@ -383,13 +470,14 @@ class WDFReader(object):
                 print("Reshaping spectra array failed. Please check.",
                       file=stderr)
             return
-        elif all(hasattr(self, "spectra_" + n) for n in ("w", "h")):
+        elif hasattr(self, "map_shape"):
             # Is a mapping
-            if self.spectra_w * self.spectra_h != self.count:
+            spectra_w, spectra_h = self.map_shape
+            if spectra_w * spectra_h != self.count:
                 print("Mapping information from WMAP not corresponding to ORGN! " +
                       "Will not reshape the spectra", file=stderr)
                 return
-            elif self.spectra_w * self.spectra_h * self.point_per_spectrum \
+            elif spectra_w * spectra_h * self.point_per_spectrum \
                     != len(self.spectra):
                 print("Mapping information from WMAP not corresponding to DATA! " +
                       "Will not reshape the spectra", file=stderr)
@@ -397,9 +485,9 @@ class WDFReader(object):
             else:
                 # Should be h rows * w columns. numpy.ndarray is row first
                 # Reshape to 3D matrix when doing 2D mapping
-                if (self.spectra_h > 1) and (self.spectra_w > 1):
+                if (spectra_h > 1) and (spectra_w > 1):
                     self.spectra = numpy.reshape(self.spectra,
-                                                 (self.spectra_h, self.spectra_w,
+                                                 (spectra_h, spectra_w,
                                                   self.point_per_spectrum))
                 # otherwise it is a line scan or series
                 else:
@@ -412,31 +500,28 @@ class WDFReader(object):
     def print_info(self):
         """Print information of the wdf file
         """
-        if not self.quiet:
-            s = []
-            s.append("{0:>24s}:\t{1}".format("Title", self.title))
-            s.append("{0:>17s} version:\t{1}.{2}.{3}.{4}".
-                     format(self.application_name,
-                            *self.application_version))
+        s = []
+        s.append("{0:>24s}:\t{1}".format("Title", self.title))
+        s.append("{0:>17s} version:\t{1}.{2}.{3}.{4}".
+                 format(self.application_name,
+                        *self.application_version))
 
-            s.append("{0:>24s}:\t{1} nm".format("Laser Wavelength",
-                                                self.laser_length))
-            for a, t in zip(["count", "capacity", "point_per_spectrum",
-                             "scan_type", "measurement_type",
-                             "spectral_units",
-                             "xlist_units", "xlist_length",
-                             "ylist_units", "ylist_length",
-                             "data_origin_count"],
-
-                            [None, None, None,
-                             ScanType, MeasurementType,
-                             UnitType,
-                             UnitType, None,
-                             UnitType, None, None, ]):
-                sname = convert_attr_name(a)
-                val = self._get_type_string(a, t)
-                s.append("{0:>24s}:\t{1}".format(sname, val))
-            print("\n".join(s))
+        s.append("{0:>24s}:\t{1} nm".format("Laser Wavelength",
+                                            self.laser_length))
+        for a in ("count", "capacity", "point_per_spectrum",
+                  "scan_type", "measurement_type",
+                  "spectral_unit",
+                  "xlist_unit", "xlist_length",
+                  "ylist_unit", "ylist_length",
+                  "xpos_unit", "ypos_unit"):
+            sname = convert_attr_name(a)
+            # Use explicit string conversion to replace
+            try:
+                val = str(getattr(self, a))
+            except AttributeError:
+                continue
+            s.append("{0:>24s}:\t{1}".format(sname, val))
+        print("\n".join(s))
 
 
 if __name__ == '__main__':
